@@ -27,21 +27,21 @@ typedef struct {
 
 typedef struct {
 	short			closed;
-	env_data*		env;			/* the DB enviroment this is in */
+	env_data*		env;			/* the DB environment this is in */
 	isc_db_handle	db;				/* the database handle */
-	char			dpb_buffer[256];/* holds the database paramet buffer */
+	char			dpb_buffer[256];/* holds the database parameter buffer */
 	short			dpb_length;		/* the used amount of the dpb */
 	isc_tr_handle	transaction;	/* the transaction handle */
 	int				lock;			/* lock count for open cursors */
-	int				autocommit;		/* should each statement be commited */
+	int				autocommit;		/* should each statement be committed */
 } conn_data;
 
 typedef struct {
 	short			closed;
-	env_data*		env;			/* the DB enviroment this is in */
+	env_data*		env;			/* the DB environment this is in */
 	conn_data*		conn;			/* the DB connection this cursor is from */
 	isc_stmt_handle stmt;			/* the statement handle */
-	int			stmt_type;			/* the type of the statment */
+	int			stmt_type;			/* the type of the statement */
 	XSQLDA			*out_sqlda;		/* the cursor data array */
 } cur_data;
 
@@ -139,7 +139,7 @@ static int cur_shut(lua_State *L, cur_data *cur)
 	/* free the cursor data */
 	free_cur(cur);
 
-	/* remove cursor from lock count and check if statment can be unregistered */
+	/* remove cursor from lock count and check if statement can be unregistered */
 	cur->closed = 1;
 	--cur->conn->lock;
 
@@ -454,7 +454,7 @@ static int conn_execute (lua_State *L) {
 	/* what do we return? a cursor or a count */
 	if(cur.out_sqlda->sqld > 0) { /* a cursor */
 		char cur_name[32];
-		cur_data* user_cur = (cur_data*)lua_newuserdata(L, sizeof(cur_data));
+		cur_data* user_cur = (cur_data*)LUASQL_NEWUD(L, sizeof(cur_data));
 		luasql_setmeta (L, LUASQL_CURSOR_FIREBIRD);
 
 		sprintf(cur_name, "dyn_cursor_%p", (void *)user_cur);
@@ -481,7 +481,7 @@ static int conn_execute (lua_State *L) {
 
 		lua_pushnumber(L, count);
 
-		/* totaly finnished with the cursor */
+		/* totally finished with the cursor */
 		isc_dsql_free_statement(conn->env->status_vector, &cur.stmt, DSQL_drop);
 		free(cur.out_sqlda);
 	}
@@ -506,7 +506,7 @@ static int conn_commit(lua_State *L) {
 /*
 ** Rolls back the current transaction
 ** Lua Returns:
-**   1 if rollback is sucsessful
+**   1 if rollback is successful
 **   nil and error message otherwise.
 */
 static int conn_rollback(lua_State *L) {
@@ -551,7 +551,8 @@ static int conn_close (lua_State *L) {
 	/* already closed */
 	if(conn->closed != 0) {
 		lua_pushboolean(L, 0);
-		return 1;
+		lua_pushstring(L, "Connection is already closed");
+		return 2;
 	}
 
 	/* are all related cursors closed? */
@@ -711,7 +712,7 @@ static void push_column(lua_State *L, int i, cur_data *cur) {
 												LUAL_BUFFERSIZE, buffer );
 			}
 
-			/* finnished, close the BLOB */
+			/* finished, close the BLOB */
 			isc_close_blob(cur->env->status_vector, &blob_handle);
 			blob_handle = 0;
 
@@ -781,7 +782,10 @@ static int cur_fetch (lua_State *L) {
 		/* close cursor for procedures/returnings as they (currently) only
 		   return one result, and error on subsequent fetches */
 		if (cur->stmt_type == isc_info_sql_stmt_exec_procedure) {
-			cur_shut(L, cur);
+			int shut_res = cur_shut(L, cur);
+			if (shut_res != 0) {
+				return shut_res;
+			}
 		}
 
 		return res;
@@ -887,19 +891,19 @@ static int cur_close (lua_State *L) {
 	luaL_argcheck (L, cur != NULL, 1, "cursor expected");
 	int shut_res;
 
-	if(cur->closed == 0) {
+	if (cur->closed == 0) {
 		shut_res = cur_shut(L, cur);
-		if(shut_res > 0) {
+		if (shut_res > 0) {
 			return shut_res;
 		}
 
-		/* return sucsess */
 		lua_pushboolean(L, 1);
 		return 1;
 	}
 
 	lua_pushboolean(L, 0);
-	return 1;
+	lua_pushstring(L, "cursor is already closed");
+	return 2;
 }
 
 /*
@@ -910,7 +914,10 @@ static int cur_gc (lua_State *L) {
 	luaL_argcheck (L, cur != NULL, 1, "cursor expected");
 
 	if(cur->closed == 0) {
-		cur_shut(L, cur);
+		int shut_res = cur_shut(L, cur);
+		if (shut_res > 0) {
+			return shut_res;
+		}
 	}
 
 	return 0;
@@ -923,7 +930,7 @@ static int create_environment (lua_State *L) {
 	int i;
 	env_data *env;
 
-	env = (env_data *)lua_newuserdata (L, sizeof (env_data));
+	env = (env_data *)LUASQL_NEWUD (L, sizeof (env_data));
 	luasql_setmeta (L, LUASQL_ENVIRONMENT_FIREBIRD);
 	/* fill in structure */
 	for(i=0; i<20; i++)
@@ -940,7 +947,7 @@ static int create_environment (lua_State *L) {
 **   source: data source
 **   user, pass: data source authentication information
 ** Lua Returns:
-**   connection object if successfull
+**   connection object if successful
 **   nil and error message otherwise.
 */
 static int env_connect (lua_State *L) {
@@ -1000,7 +1007,7 @@ static int env_connect (lua_State *L) {
 		return return_db_error(L, conn.env->status_vector);
 
 	/* create the lua object and add the connection to the lock */
-	res_conn = (conn_data*)lua_newuserdata(L, sizeof(conn_data));
+	res_conn = (conn_data*)LUASQL_NEWUD(L, sizeof(conn_data));
 	luasql_setmeta (L, LUASQL_CONNECTION_FIREBIRD);
 	memcpy(res_conn, &conn, sizeof(conn_data));
 	res_conn->closed = 0;   /* connect now officially open */
@@ -1023,14 +1030,22 @@ static int env_close (lua_State *L) {
 	luaL_argcheck (L, env != NULL, 1, "environment expected");
 	
 	/* already closed? */
-	if(env->closed == 1) {
+	if(env->closed) {
 		lua_pushboolean(L, 0);
-		return 1;
+		lua_pushstring(L, "env is already closed");
+		return 2;
 	}
 
 	/* check the lock */
+/*
 	if(env->lock > 0)
 		return luasql_faildirect(L, "there are still open connections");
+*/
+	if(env->lock > 0) {
+		lua_pushboolean(L, 0);
+		lua_pushstring(L, "there are open connections");
+		return 2;
+	}
 
 	/* unregister */
 	lua_unregisterobj(L, env);
@@ -1046,7 +1061,7 @@ static int env_close (lua_State *L) {
 ** GCs an environment object
 */
 static int env_gc (lua_State *L) {
-	/* nothing to be done for the FB envronment */
+	/* nothing to be done for the FB environment */
 	return 0;
 }
 
@@ -1056,12 +1071,14 @@ static int env_gc (lua_State *L) {
 static void create_metatables (lua_State *L) {
 	struct luaL_Reg environment_methods[] = {
 		{"__gc", env_gc},
+		{"__close", env_gc},
 		{"close", env_close},
 		{"connect", env_connect},
 		{NULL, NULL},
 	};
 	struct luaL_Reg connection_methods[] = {
 		{"__gc", conn_gc},
+		{"__close", conn_gc},
 		{"close", conn_close},
 		{"execute", conn_execute},
 		{"commit", conn_commit},
@@ -1072,6 +1089,7 @@ static void create_metatables (lua_State *L) {
 	};
 	struct luaL_Reg cursor_methods[] = {
 		{"__gc", cur_gc},
+		{"__close", cur_gc},
 		{"close", cur_close},
 		{"fetch", cur_fetch},
 		{"getcoltypes", cur_coltypes},
