@@ -24,7 +24,6 @@
 #define LUASQL_CONNECTION_OCI8 "Oracle connection"
 #define LUASQL_CURSOR_OCI8 "Oracle cursor"
 
-
 typedef struct {
 	short         closed;
 	int           conn_counter;
@@ -269,7 +268,7 @@ static int free_column_buffers (lua_State *L, cur_data *cur, int i) {
 			break;
 		default:
 			luaL_error (L, LUASQL_PREFIX"unknown type");
-			/*printf("free_buffers(): Unknow Type: %d count: %d\n",cols.item[count].type, count );*/
+			/*printf("free_buffers(): Unknown Type: %d count: %d\n",cols.item[count].type, count );*/
 			break;
 	}
     return 0;
@@ -400,7 +399,8 @@ static int cur_close (lua_State *L) {
 	luaL_argcheck (L, cur != NULL, 1, LUASQL_PREFIX"cursor expected");
 	if (cur->closed) {
 		lua_pushboolean (L, 0);
-		return 1;
+		lua_pushstring (L, "Cursor is already closed");
+		return 2;
 	}
 
 	/* Deallocate buffers. */
@@ -522,10 +522,14 @@ static int conn_close (lua_State *L) {
 	luaL_argcheck (L, conn != NULL, 1, LUASQL_PREFIX"connection expected");
 	if (conn->closed) {
 		lua_pushboolean (L, 0);
-		return 1;
+		lua_pushstring (L, "Connection is already closed");
+		return 2;
+	}
+	if (conn->cur_counter > 0){
+		lua_pushboolean (L, 0);
+		lua_pushstring (L, "There are open cursors");
+		return 2;
 	}
-	if (conn->cur_counter > 0)
-		return luaL_error (L, LUASQL_PREFIX"there are open cursors");
 
 	/* Nullify structure fields. */
 	conn->closed = 1;
@@ -554,7 +558,7 @@ static int conn_close (lua_State *L) {
 static int create_cursor (lua_State *L, int o, conn_data *conn, OCIStmt *stmt, const char *text) {
 	int i;
 	env_data *env;
-	cur_data *cur = (cur_data *)lua_newuserdata(L, sizeof(cur_data));
+	cur_data *cur = (cur_data *)LUASQL_NEWUD(L, sizeof(cur_data));
 	luasql_setmeta (L, LUASQL_CURSOR_OCI8);
 
 	conn->cur_counter++;
@@ -726,7 +730,7 @@ static int env_connect (lua_State *L) {
 	size_t userlen = (username) ? strlen(username) : 0;
 	size_t passlen = (password) ? strlen(password) : 0;
 	/* Alloc connection object */
-	conn_data *conn = (conn_data *)lua_newuserdata(L, sizeof(conn_data));
+	conn_data *conn = (conn_data *)LUASQL_NEWUD(L, sizeof(conn_data));
 
 	/* fill in structure */
 	luasql_setmeta (L, LUASQL_CONNECTION_OCI8);
@@ -770,13 +774,17 @@ static int env_close (lua_State *L) {
 	luaL_argcheck (L, env != NULL, 1, LUASQL_PREFIX"environment expected");
 	if (env->closed) {
 		lua_pushboolean (L, 0);
-		return 1;
+		lua_pushstring (L, "Environment is already closed");
+		return 2;
+	}
+	if (env->conn_counter > 0){
+		lua_pushboolean (L, 0);
+		lua_pushstring (L, "There are open connections");
+		return 2;
 	}
-	if (env->conn_counter > 0)
-		return luaL_error (L, LUASQL_PREFIX"there are open connections");
 
 	env->closed = 1;
-	/* desalocar: env->errhp e env->envhp */
+	/* release resources */
 	if (env->envhp)
 		OCIHandleFree ((dvoid *)env->envhp, OCI_HTYPE_ENV);
 	if (env->errhp)
@@ -790,7 +798,7 @@ static int env_close (lua_State *L) {
 ** Creates an Environment and returns it.
 */
 static int create_environment (lua_State *L) {
-	env_data *env = (env_data *)lua_newuserdata(L, sizeof(env_data));
+	env_data *env = (env_data *)LUASQL_NEWUD(L, sizeof(env_data));
 	luasql_setmeta (L, LUASQL_ENVIRONMENT_OCI8);
 	/* fill in structure */
 	env->closed = 0;
@@ -819,12 +827,14 @@ static int create_environment (lua_State *L) {
 static void create_metatables (lua_State *L) {
 	struct luaL_Reg environment_methods[] = {
 		{"__gc", env_close}, /* Should this method be changed? */
+		{"__close", env_close},
 		{"close", env_close},
 		{"connect", env_connect},
 		{NULL, NULL},
 	};
 	struct luaL_Reg connection_methods[] = {
 		{"__gc", conn_close}, /* Should this method be changed? */
+		{"__close", conn_close},
 		{"close", conn_close},
 		{"execute", conn_execute},
 		{"commit", conn_commit},
@@ -834,6 +844,7 @@ static void create_metatables (lua_State *L) {
 	};
 	struct luaL_Reg cursor_methods[] = {
 		{"__gc", cur_close}, /* Should this method be changed? */
+		{"__close", cur_close},
 		{"close", cur_close},
 		{"getcolnames", cur_getcolnames},
 		{"getcoltypes", cur_getcoltypes},
