@@ -1,6 +1,5 @@
 ---------------------------------------------------------------------
 -- SQLite specific tests and configurations.
--- $Id: sqlite.lua,v 1.2 2006/05/31 21:43:33 carregal Exp $
 ---------------------------------------------------------------------
 
 ---------------------------------------------------------------------
@@ -14,4 +13,11 @@ end
 
 function checkUnknownDatabase(ENV)
 	-- skip this test
-end
\ No newline at end of file
+end
+
+function finalization ()
+	os.execute ("rm -rf "..datasource)
+end
+
+table.insert (CONN_METHODS, "escape")
+table.insert (EXTENSIONS, escape)
