@@ -1,5 +1,6 @@
-#!/usr/local/bin/lua50
+#!/usr/local/bin/lua5.1
 -- See Copyright Notice in license.html
+-- $Id: test.lua,v 1.52 2008/06/30 10:43:03 blumf Exp $
 
 TOTAL_FIELDS = 40
 TOTAL_ROWS = 40 --unused
@@ -10,6 +11,33 @@ QUERYING_STRING_TYPE_NAME = "text"
 CREATE_TABLE_RETURN_VALUE = 0
 DROP_TABLE_RETURN_VALUE = 0
 
+MSG_CURSOR_NOT_CLOSED = "cursor was not automatically closed by fetch"
+
+CHECK_GETCOL_INFO_TABLES = true
+
+---------------------------------------------------------------------
+if not string.find(_VERSION, " 5.0") then
+	table.getn = assert((loadstring or load)[[return function (t) return #t end]])()
+end
+
+---------------------------------------------------------------------
+-- Creates a table that can handle differing capitalization of field
+-- names
+-- @return A table with altered metatable
+---------------------------------------------------------------------
+local mt = {
+	__index = function(t, i)
+		if type(i) == "string" then
+			return rawget(t, string.upper(i)) or rawget(t, string.lower(i))
+		end
+
+		return rawget(t, i)
+	end
+}
+function fetch_table ()
+	return setmetatable({}, mt)
+end
+
 ---------------------------------------------------------------------
 -- Produces a SQL statement which completely erases a table.
 -- @param table_name String with the name of the table.
@@ -33,12 +61,30 @@ function assert2 (expected, value, msg)
 		tostring(expected)..")")
 end
 
+---------------------------------------------------------------------
+-- Shallow compare of two tables
+---------------------------------------------------------------------
+function table_compare(t1, t2)
+	if t1 == t2 then return true; end
+
+	for i, v in pairs(t1) do
+		if t2[i] ~= v then return false; end
+	end
+
+	for i, v in pairs(t2) do
+		if t1[i] ~= v then return false; end
+	end
+
+	return true
+end
+
 ---------------------------------------------------------------------
 -- object test.
 ---------------------------------------------------------------------
 function test_object (obj, objmethods)
 	-- checking object type.
 	assert2 (true, type(obj) == "userdata" or type(obj) == "table", "incorrect object type")
+
 	-- trying to get metatable.
 	assert2 ("LuaSQL: you're not allowed to get this metatable",
 		getmetatable(obj), "error permitting access to object's metatable")
@@ -84,6 +130,7 @@ function basic_test ()
 	assert2 (false, ENV:close())
 	-- Reopen the environment.
 	ENV = ENV_OK (luasql[driver] ())
+
 	-- Check connection object.
 	local conn, err = ENV:connect (datasource, username, password)
 	assert (conn, (err or '').." ("..datasource..")")
@@ -91,7 +138,7 @@ function basic_test ()
 	assert2 (true, conn:close(), "couldn't close connection")
 	-- trying to execute a statement with a closed connection.
 	assert2 (false, pcall (conn.execute, conn, "create table x (c char)"),
-		"error connecting with a closed environment")
+		"error while executing through a closed connection")
 	-- it is ok to close a closed object, but false is returned instead of true.
 	assert2 (false, conn:close())
 	-- Check error situation.
@@ -126,6 +173,7 @@ end
 function create_table ()
 	-- Check SQL statements.
 	CONN = CONN_OK (ENV:connect (datasource, username, password))
+	CONN:execute"drop table t"
 	-- Create t.
 	local cmd = define_table(TOTAL_FIELDS)
 	assert2 (CREATE_TABLE_RETURN_VALUE, CONN:execute (cmd))
@@ -135,6 +183,9 @@ end
 -- Fetch 2 values.
 ---------------------------------------------------------------------
 function fetch2 ()
+	-- check number of lines
+	local cur0 = CUR_OK (CONN:execute ("select count(*) from t"))
+	assert2 (0, tonumber (cur0:fetch()))
 	-- insert a record.
 	assert2 (1, CONN:execute ("insert into t (f1, f2) values ('b', 'c')"))
 	-- retrieve data.
@@ -145,7 +196,7 @@ function fetch2 ()
 	assert2 ('c', f2)
 	assert2 (nil, f3)
 	assert2 (nil, cur:fetch())
-	assert2 (true, cur:close(), "couldn't close cursor")
+	assert2 (false, cur:close(), MSG_CURSOR_NOT_CLOSED)
 	assert2 (false, cur:close())
 	-- insert a second record.
 	assert2 (1, CONN:execute ("insert into t (f1, f2) values ('d', 'e')"))
@@ -159,10 +210,13 @@ function fetch2 ()
 	assert2 ('e', f2)
 	assert2 (nil, f3)
 	assert2 (nil, cur:fetch())
-	assert2 (true, cur:close(), "couldn't close cursor")
+	assert2 (false, cur:close(), MSG_CURSOR_NOT_CLOSED)
 	assert2 (false, cur:close())
 	-- remove records.
 	assert2 (2, CONN:execute ("delete from t where f1 in ('b', 'd')"))
+	-- check number of lines
+	local cur0 = CUR_OK (CONN:execute ("select count(*) from t"))
+	assert2 (0, tonumber (cur0:fetch()))
 end
 
 ---------------------------------------------------------------------
@@ -170,12 +224,15 @@ end
 -- indexing.
 ---------------------------------------------------------------------
 function fetch_new_table ()
+	-- check number of lines
+	local cur0 = CUR_OK (CONN:execute ("select count(*) from t"))
+	assert2 (0, tonumber (cur0:fetch()))
 	-- insert elements.
 	assert2 (1, CONN:execute ("insert into t (f1, f2, f3, f4) values ('a', 'b', 'c', 'd')"))
 	assert2 (1, CONN:execute ("insert into t (f1, f2, f3, f4) values ('f', 'g', 'h', 'i')"))
 	-- retrieve data using a new table.
 	local cur = CUR_OK (CONN:execute ("select f1, f2, f3, f4 from t order by f1"))
-	local row, err = cur:fetch{}
+	local row, err = cur:fetch(fetch_table())
 	assert2 (type(row), "table", err)
 	assert2 ('a', row[1])
 	assert2 ('b', row[2])
@@ -185,35 +242,8 @@ function fetch_new_table ()
 	assert2 (nil, row.f2)
 	assert2 (nil, row.f3)
 	assert2 (nil, row.f4)
-	row, err = cur:fetch{}
-	assert (type(row), "table", err)
-	assert2 ('f', row[1])
-	assert2 ('g', row[2])
-	assert2 ('h', row[3])
-	assert2 ('i', row[4])
-	assert2 (nil, row.f1)
-	assert2 (nil, row.f2)
-	assert2 (nil, row.f3)
-	assert2 (nil, row.f4)
-	assert2 (nil, cur:fetch())
-	assert2 (true, cur:close(), "couldn't close cursor")
-	assert2 (false, cur:close())
-
-	-- retrieve data reusing the same table.
-	io.write ("reusing a table...")
-	cur = CUR_OK (CONN:execute ("select f1, f2, f3, f4 from t order by f1"))
-	local row, err = cur:fetch{}
-	assert (type(row), "table", err)
-	assert2 ('a', row[1])
-	assert2 ('b', row[2])
-	assert2 ('c', row[3])
-	assert2 ('d', row[4])
-	assert2 (nil, row.f1)
-	assert2 (nil, row.f2)
-	assert2 (nil, row.f3)
-	assert2 (nil, row.f4)
-	row, err = cur:fetch (row)
-	assert (type(row), "table", err)
+	row, err = cur:fetch(fetch_table())
+	assert2 (type(row), "table", err)
 	assert2 ('f', row[1])
 	assert2 ('g', row[2])
 	assert2 ('h', row[3])
@@ -223,14 +253,41 @@ function fetch_new_table ()
 	assert2 (nil, row.f3)
 	assert2 (nil, row.f4)
 	assert2 (nil, cur:fetch{})
-	assert2 (true, cur:close(), "couldn't close cursor")
+	assert2 (false, cur:close(), MSG_CURSOR_NOT_CLOSED)
+	assert2 (false, cur:close())
+
+	-- retrieve data reusing the same table.
+	io.write ("reusing a table...")
+	cur = CUR_OK (CONN:execute ("select f1, f2, f3, f4 from t order by f1"))
+	local row, err = cur:fetch(fetch_table())
+	assert2 (type(row), "table", err)
+	assert2 ('a', row[1])
+	assert2 ('b', row[2])
+	assert2 ('c', row[3])
+	assert2 ('d', row[4])
+	assert2 (nil, row.f1)
+	assert2 (nil, row.f2)
+	assert2 (nil, row.f3)
+	assert2 (nil, row.f4)
+	row, err = cur:fetch (row)
+	assert2 (type(row), "table", err)
+	assert2 ('f', row[1])
+	assert2 ('g', row[2])
+	assert2 ('h', row[3])
+	assert2 ('i', row[4])
+	assert2 (nil, row.f1)
+	assert2 (nil, row.f2)
+	assert2 (nil, row.f3)
+	assert2 (nil, row.f4)
+	assert2 (nil, cur:fetch(fetch_table()))
+	assert2 (false, cur:close(), MSG_CURSOR_NOT_CLOSED)
 	assert2 (false, cur:close())
 
 	-- retrieve data reusing the same table with alphabetic indexes.
 	io.write ("with alpha keys...")
 	cur = CUR_OK (CONN:execute ("select f1, f2, f3, f4 from t order by f1"))
-	local row, err = cur:fetch ({}, "a")
-	assert (type(row), "table", err)
+	local row, err = cur:fetch (fetch_table(), "a")
+	assert2 (type(row), "table", err)
 	assert2 (nil, row[1])
 	assert2 (nil, row[2])
 	assert2 (nil, row[3])
@@ -250,14 +307,14 @@ function fetch_new_table ()
 	assert2 ('h', row.f3)
 	assert2 ('i', row.f4)
 	assert2 (nil, cur:fetch(row, "a"))
-	assert2 (true, cur:close(), "couldn't close cursor")
+	assert2 (false, cur:close(), MSG_CURSOR_NOT_CLOSED)
 	assert2 (false, cur:close())
 
 	-- retrieve data reusing the same table with both indexes.
 	io.write ("with both keys...")
 	cur = CUR_OK (CONN:execute ("select f1, f2, f3, f4 from t order by f1"))
-	local row, err = cur:fetch ({}, "an")
-	assert (type(row), "table", err)
+	local row, err = cur:fetch (fetch_table(), "an")
+	assert2 (type(row), "table", err)
 	assert2 ('a', row[1])
 	assert2 ('b', row[2])
 	assert2 ('c', row[3])
@@ -267,7 +324,7 @@ function fetch_new_table ()
 	assert2 ('c', row.f3)
 	assert2 ('d', row.f4)
 	row, err = cur:fetch (row, "an")
-	assert (type(row), "table", err)
+	assert2 (type(row), "table", err)
 	assert2 ('f', row[1])
 	assert2 ('g', row[2])
 	assert2 ('h', row[3])
@@ -277,16 +334,22 @@ function fetch_new_table ()
 	assert2 ('h', row.f3)
 	assert2 ('i', row.f4)
 	assert2 (nil, cur:fetch(row, "an"))
-	assert2 (true, cur:close(), "couldn't close cursor")
+	assert2 (false, cur:close(), MSG_CURSOR_NOT_CLOSED)
 	assert2 (false, cur:close())
 	-- clean the table.
 	assert2 (2, CONN:execute ("delete from t where f1 in ('a', 'f')"))
+	-- check number of lines
+	local cur0 = CUR_OK (CONN:execute ("select count(*) from t"))
+	assert2 (0, tonumber (cur0:fetch()))
 end
 
 ---------------------------------------------------------------------
 -- Fetch many values
 ---------------------------------------------------------------------
 function fetch_many ()
+	-- check number of lines
+	local cur0 = CUR_OK (CONN:execute ("select count(*) from t"))
+	assert2 (0, tonumber (cur0:fetch()))
 	-- insert values.
 	local fields, values = "f1", "'v1'"
 	for i = 2, TOTAL_FIELDS do
@@ -304,45 +367,41 @@ function fetch_many ()
 		assert2 ('v'..i, row[i])
 	end
 	assert2 (nil, cur:fetch (row))
-	assert2 (true, cur:close(), "couldn't close cursor")
-	assert2 (false, cur:close())
+	assert2 (false, cur:close(), MSG_CURSOR_NOT_CLOSED)
 	-- fetch values (with a table and default indexing).
 	io.write ("with a table...")
 	local cur = CUR_OK (CONN:execute ("select * from t where f1 = 'v1'"))
-	local row = cur:fetch {}
+	local row = cur:fetch(fetch_table())
 	assert2 ("string", type(row[1]), "error while trying to fetch many values (default indexing)")
 	for i = 1, TOTAL_FIELDS do
 		assert2 ('v'..i, row[i])
 	end
 	assert2 (nil, cur:fetch (row))
-	assert2 (true, cur:close(), "couldn't close cursor")
-	assert2 (false, cur:close())
+	assert2 (false, cur:close(), MSG_CURSOR_NOT_CLOSED)
 	-- fetch values (with numbered indexes on a table).
 	io.write ("with numbered keys...")
 	local cur = CUR_OK (CONN:execute ("select * from t where f1 = 'v1'"))
-	local row = cur:fetch ({}, "n")
+	local row = cur:fetch (fetch_table(), "n")
 	assert2 ("string", type(row[1]), "error while trying to fetch many values (numbered indexes)")
 	for i = 1, TOTAL_FIELDS do
 		assert2 ('v'..i, row[i])
 	end
 	assert2 (nil, cur:fetch (row))
-	assert2 (true, cur:close(), "couldn't close cursor")
-	assert2 (false, cur:close())
+	assert2 (false, cur:close(), MSG_CURSOR_NOT_CLOSED)
 	-- fetch values (with alphanumeric indexes on a table).
 	io.write ("with alpha keys...")
 	local cur = CUR_OK (CONN:execute ("select * from t where f1 = 'v1'"))
-	local row = cur:fetch ({}, "a")
+	local row = cur:fetch (fetch_table(), "a")
 	assert2 ("string", type(row.f1), "error while trying to fetch many values (alphanumeric indexes)")
 	for i = 1, TOTAL_FIELDS do
 		assert2 ('v'..i, row['f'..i])
 	end
 	assert2 (nil, cur:fetch (row))
-	assert2 (true, cur:close(), "couldn't close cursor")
-	assert2 (false, cur:close())
+	assert2 (false, cur:close(), MSG_CURSOR_NOT_CLOSED)
 	-- fetch values (with both indexes on a table).
 	io.write ("with both keys...")
 	local cur = CUR_OK (CONN:execute ("select * from t where f1 = 'v1'"))
-	local row = cur:fetch ({}, "na")
+	local row = cur:fetch (fetch_table(), "na")
 	assert2 ("string", type(row[1]), "error while trying to fetch many values (both indexes)")
 	assert2 ("string", type(row.f1), "error while trying to fetch many values (both indexes)")
 	for i = 1, TOTAL_FIELDS do
@@ -350,30 +409,36 @@ function fetch_many ()
 		assert2 ('v'..i, row['f'..i])
 	end
 	assert2 (nil, cur:fetch (row))
-	assert2 (true, cur:close(), "couldn't close cursor")
-	assert2 (false, cur:close())
+	assert2 (false, cur:close(), MSG_CURSOR_NOT_CLOSED)
 	-- clean the table.
 	assert2 (1, CONN:execute ("delete from t where f1 = 'v1'"))
+	-- check number of lines
+	local cur0 = CUR_OK (CONN:execute ("select count(*) from t"))
+	assert2 (0, tonumber (cur0:fetch()))
 end
 
 ---------------------------------------------------------------------
 ---------------------------------------------------------------------
 function rollback ()
-	CONN:setautocommit (false) -- == begin transaction
+	-- check number of lines
+	local cur0 = CUR_OK (CONN:execute ("select count(*) from t"))
+	assert2 (0, tonumber (cur0:fetch()))
+	-- begin transaction
+	assert2 (true, CONN:setautocommit (false), "couldn't disable autocommit")
 	-- insert a record and commit the operation.
 	assert2 (1, CONN:execute ("insert into t (f1) values ('a')"))
 	local cur = CUR_OK (CONN:execute ("select count(*) from t"))
 	assert2 (1, tonumber (cur:fetch ()), "Insert failed")
 	assert2 (true, cur:close(), "couldn't close cursor")
 	assert2 (false, cur:close())
-	CONN:commit ()
+	assert2 (true, CONN:commit(), "couldn't commit transaction")
 	-- insert a record and roll back the operation.
 	assert2 (1, CONN:execute ("insert into t (f1) values ('b')"))
 	local cur = CUR_OK (CONN:execute ("select count(*) from t"))
 	assert2 (2, tonumber (cur:fetch ()), "Insert failed")
 	assert2 (true, cur:close(), "couldn't close cursor")
 	assert2 (false, cur:close())
-	CONN:rollback ()
+	assert2 (true, CONN:rollback (), "couldn't roolback transaction")
 	-- check resulting table with one record.
 	cur = CUR_OK (CONN:execute ("select count(*) from t"))
 	assert2 (1, tonumber(cur:fetch()), "Rollback failed")
@@ -385,7 +450,7 @@ function rollback ()
 	assert2 (0, tonumber(cur:fetch()))
 	assert2 (true, cur:close(), "couldn't close cursor")
 	assert2 (false, cur:close())
-	CONN:rollback ()
+	assert2 (true, CONN:rollback (), "couldn't roolback transaction")
 	-- check resulting table with one record.
 	cur = CUR_OK (CONN:execute ("select count(*) from t"))
 	assert2 (1, tonumber(cur:fetch()), "Rollback failed")
@@ -400,7 +465,7 @@ function rollback ()
 	assert2 (2, tonumber (cur:fetch ()), "Insert failed")
 	assert2 (true, cur:close(), "couldn't close cursor")
 	assert2 (false, cur:close())
-	CONN:setautocommit (true)
+	assert2 (true, CONN:setautocommit (true), "couldn't enable autocommit")
 	-- check resulting table with one record.
 	cur = CUR_OK (CONN:execute ("select count(*) from t"))
 	assert2 (1, tonumber(cur:fetch()), "Rollback failed")
@@ -409,8 +474,8 @@ function rollback ()
 --]]
 	-- clean the table.
 	assert2 (1, CONN:execute (sql_erase_table"t"))
-	CONN:commit ()
-	CONN:setautocommit (true)
+	assert2 (true, CONN:commit (), "couldn't commit transaction")
+	assert2 (true, CONN:setautocommit (true), "couldn't enable autocommit")
 	-- check resulting table with no records.
 	cur = CUR_OK (CONN:execute ("select count(*) from t"))
 	assert2 (0, tonumber(cur:fetch()), "Rollback failed")
@@ -432,14 +497,20 @@ function column_info ()
 	assert2 (4, table.getn(names), "incorrect column names table")
 	assert2 (4, table.getn(types), "incorrect column types table")
 	for i = 1, table.getn(names) do
-		assert2 ("f"..i, names[i], "incorrect column names table")
+		assert2 ("f"..i, string.lower(names[i]), "incorrect column names table")
 		local type_i = types[i]
-		assert (type_i == QUERYING_STRING_TYPE_NAME, "incorrect column types table")
+		type_i = string.lower(type_i)
+		assert2 (QUERYING_STRING_TYPE_NAME, type_i, "incorrect column types table")
 	end
 	-- check if the tables are being reused.
 	local n2, t2 = cur:getcolnames(), cur:getcoltypes()
-	assert2 (names, n2, "getcolnames is rebuilding the table")
-	assert2 (types, t2, "getcoltypes is rebuilding the table")
+	if CHECK_GETCOL_INFO_TABLES then 
+		assert2 (names, n2, "getcolnames is rebuilding the table")
+		assert2 (types, t2, "getcoltypes is rebuilding the table")
+	else
+		assert2 (true, table_compare(names, n2), "getcolnames is inconsistent")
+		assert2 (true, table_compare(types, t2), "getcoltypes is inconsistent")
+	end
 	assert2 (true, cur:close(), "couldn't close cursor")
 	assert2 (false, cur:close())
 	-- clean the table.
@@ -447,8 +518,30 @@ function column_info ()
 end
 
 ---------------------------------------------------------------------
+-- Escaping strings
+---------------------------------------------------------------------
+function escape ()
+	local escaped = CONN:escape"a'b'c'd"
+	assert ("a\\'b\\'c\\'d" == escaped or "a''b''c''d" == escaped)
+
+	local n = 5200
+	local s1 = string.rep("'", n)
+	local s2 = CONN:escape(s1)
+	local s3 = s1:gsub ("'", "\\'")
+	assert2 (n, s1:len(), "incorrect escaping of '"..s1.."': expected "..n.." bytes, but got "..s1:len())
+	assert2 (2*n, s2:len(), "incerrect escaping of '"..s2.."': expected "..(2*n).." bytes, but got "..s2:len())
+	assert (s2 == s1..s1 or s2 == s3)
+
+	io.write (" escape")
+end
+
+---------------------------------------------------------------------
+-- Check closing of various objects.
 ---------------------------------------------------------------------
 function check_close()
+	local cur0 = CUR_OK(CONN:execute"select count(*) from t")
+	assert2 (0, tonumber(cur0:fetch()))
+	cur0:close()
 	-- an object with references to it can't be closed
 	local cmd = "select * from t"
 	local cur = CUR_OK(CONN:execute (cmd))
@@ -472,30 +565,84 @@ function check_close()
 	collectgarbage ()
 	assert2(nil, a.CONN, "connection not collected")
 
-	-- check cursor integrity after trying to close a connection
+	-- check cursor integrity after trying to close its connection
 	local conn = CONN_OK (ENV:connect (datasource, username, password))
 	assert2 (1, conn:execute"insert into t (f1) values (1)", "could not insert a new record")
 	local cur = CUR_OK (conn:execute (cmd))
-	local ok, err = pcall (conn.close, conn)
+	local ok, err, msg = pcall (conn.close, conn)
+	assert2 (true, ok, "couldn´t try to close the connection")
+	-- err could be true if the driver DOESN'T care about closing a connection
+	--	that has open cursors
+	-- err could be false if the driver DOES care about closing a connection
+	--	that has open cursors
 	CUR_OK (cur)
 	assert (cur:fetch(), "corrupted cursor")
 	cur:close ()
 	conn:close ()
+	-- The following check is failing in Firebird
+    --local cur0 = CUR_OK (CONN:execute ("select count(*) from t"))
+    --assert2 (1, tonumber (cur0:fetch()))
+	--cur0:close()
+
+	-- check connection integrity after trying to close an environment
+	local conn = CONN_OK (ENV:connect (datasource, username, password))
+	local closed = ENV:close()
+	--assert2 (true, ENV:close(), "couldn't close the environment!")
+	if closed then
+		-- Some drivers can close the environment with open connections
+		CONN_OK (conn)
+		local cmd = "select * from t"
+		local cur = CUR_OK(CONN:execute (cmd))
+		assert2 ('1', cur:fetch(), "couldn't retrieve a row from `t`")
+		assert2 (true, cur:close(), "couldn't close the cursor!")
+    local cur0 = CUR_OK (CONN:execute ("select count(*) from t"))
+    assert2 (1, tonumber (cur0:fetch()))
+	cur0:close()
+
+		assert2 (true, conn:close(), "couldn't close the connection!")
+		ENV = ENV_OK (luasql[driver] ())
+	else
+		-- Some drivers cannot close the environment with open connections
+		conn:close()
+		ENV:close()
+	end
+end
+
+---------------------------------------------------------------------
+-- Check support for to-be-closed variables.
+-- Since this feature depends on a syntax construction not available
+-- in versions prior to 5.4, this check should be written in another
+-- file (or in a string).
+---------------------------------------------------------------------
+function to_be_closed_support ()
+	assert (loadfile"to_be_closed_support.lua")()
 end
 
 ---------------------------------------------------------------------
 ---------------------------------------------------------------------
 function drop_table ()
-	-- Postgres retorna 0, enquanto ODBC retorna -1.
-	CONN:setautocommit(true)
+	-- check number of lines
+	-- The following check is failing with Firebird
+	--local cur0 = CUR_OK (CONN:execute ("select count(*) from t"))
+	--assert2 (1, tonumber (cur0:fetch()))
+	--assert2 (true, cur0:close(), "couldn't close the cursor")
+
+	assert2 (true, CONN:setautocommit(true), "couldn't enable autocommit")
+	-- Postgres retorns 0, ODBC retorns -1, sqlite returns 1
 	assert2 (DROP_TABLE_RETURN_VALUE, CONN:execute ("drop table t"))
 end
 
 ---------------------------------------------------------------------
 ---------------------------------------------------------------------
 function close_conn ()
-	assert (true, CONN:close())
-	assert (true, ENV:close())
+	assert2 (true, CONN:close())
+	assert2 (true, ENV:close())
+end
+
+---------------------------------------------------------------------
+---------------------------------------------------------------------
+function finalization ()
+	-- nothing to do
 end
 
 ---------------------------------------------------------------------
@@ -504,9 +651,17 @@ end
 EXTENSIONS = {
 }
 function extensions_test ()
+	-- check number of lines
+	local cur0 = CUR_OK (CONN:execute ("select count(*) from t"))
+	assert2 (0, tonumber (cur0:fetch()))
+
 	for i, f in ipairs (EXTENSIONS) do
 		f ()
 	end
+
+	-- check number of lines
+	local cur0 = CUR_OK (CONN:execute ("select count(*) from t"))
+	assert2 (0, tonumber (cur0:fetch()))
 end
 
 ---------------------------------------------------------------------
@@ -559,13 +714,10 @@ if type(arg[1]) ~= "string" then
 end
 
 driver = arg[1]
-datasource = arg[2] or "luasql-test"
-username = arg[3] or nil
-password = arg[4] or nil
 
 -- Loading driver specific functions
 if arg[0] then
-	local path = string.gsub (arg[0], "^([^/]*%/).*$", "%1")
+	local path = string.gsub (arg[0], "^(.*%/)[^/]*$", "%1")
 	if path == "test.lua" then
 		path = ""
 	end
@@ -575,10 +727,15 @@ if arg[0] then
 		print ("LuaSQL test: couldn't find driver-specific test file ("..
 			file..").\nProceeding with general test")
 	else
+		print ("Loading driver-specific test file ("..file..").")
 		f ()
 	end
 end
 
+datasource = arg[2] or DEFAULT_TEST_DATABASE or "luasql-test"
+username = arg[3] or DEFAULT_USERNAME or nil
+password = arg[4] or DEFAULT_PASSWORD or nil
+
 -- Complete set of tests
 tests = {
 	{ "basic checking", basic_test },
@@ -590,17 +747,54 @@ tests = {
 	{ "get column information", column_info },
 	{ "extensions", extensions_test },
 	{ "close objects", check_close },
+-- to-be-closed variables could be inserted here
 	{ "drop table", drop_table },
 	{ "close connection", close_conn },
+	{ "finalization", finalization },
 }
 
-require ("luasql."..driver)
-assert (luasql, "no luasql table")
+if string.find(_VERSION, " 5.0") then
+	local init_so, err = loadlib("./"..driver..".so", "luaopen_luasql_"..driver)
+	if init_so then
+		luasql = init_so()
+	else
+		luasql = assert(loadlib("/usr/local/lib/lua/5.0/luasql/"..driver..".so", "luaopen_luasql_"..driver))()
+	end
+else
+	luasql = require ("luasql."..driver)
+	if string.find(_VERSION, " 5.4") then
+		table.insert (tests, 10, { "to-be-closed support", to_be_closed_support })
+	end
+end
+assert (luasql, "Could not load driver: no luasql table.")
+io.write (luasql._VERSION.." "..driver)
+if luasql._CLIENTVERSION then
+	io.write (" ("..luasql._CLIENTVERSION..")")
+end
+io.write (" driver test.  "..luasql._COPYRIGHT.."\n")
 
 for i = 1, table.getn (tests) do
 	local t = tests[i]
 	io.write (t[1].." ...")
-	t[2] ()
+	local ok, err = xpcall (t[2], debug.traceback)
+	if not ok then
+		io.write ("\n"..err)
+		io.write"\n... trying to drop test table ..."
+		local ok, err = pcall (drop_table)
+		if not ok then
+			io.write (" failed: "..err)
+		else
+			io.write" OK !\n... and to close the connection ..."
+			local ok, err = pcall (close_conn)
+			if not ok then
+				io.write (" failed: "..err)
+			else
+				io.write" OK !"
+			end
+		end
+		io.write"\nThe test failed!\n"
+		return
+	end
 	io.write (" OK !\n")
 end
-
+io.write ("The test passed!\n")
